/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "libfawk.h"
#include "array.h"
#include "parser.h"

FAWK_API fawk_cell_t *fawk_sym_lookup(fawk_ctx_t *ctx, const char *name)
{
	return htsp_get(&ctx->symtab, name);
}

#define FAWK_SYMTAB_REG(typ, symname, funcname, err_ret) \
	f = fawk_malloc(ctx, sizeof(fawk_cell_t)); if (f == NULL) return err_ret; \
	goto setup; setup:; \
	f->type = typ; \
	f->name = fawk_strdup(ctx, symname); if (f->name == NULL) { fawk_free(ctx, f); return err_ret; } \
	f->data.func.cfunc = NULL; \
	f->data.func.name = funcname; \
	htsp_set(&ctx->symtab, f->name, f);


FAWK_API int fawk_symtab_regcfunc(fawk_ctx_t *ctx, const char *name, fawk_cfunc_t cfunc)
{
	fawk_cell_t *f = htsp_get(&ctx->symtab, name);
	if (f != NULL)
		return -1;
	FAWK_SYMTAB_REG(FAWK_FUNC, name, f->name, -1);
	f->data.func.cfunc = cfunc;
	return 0;
}

FAWK_API int fawk_symtab_regfunc(fawk_ctx_t *ctx, const char *name, size_t addr, int numargs, int numfixedargs)
{
	fawk_cell_t *f = htsp_get(&ctx->symtab, name);
	if (f == NULL) {
		FAWK_SYMTAB_REG(FAWK_FUNC, name, f->name, -1);
		f->data.func.ip = addr;
		f->data.func.numargs = numargs;
		f->data.func.numfixedargs = numfixedargs;
	}
	else { /* update existing entry if it was a forward declaration */
		if (f->type != FAWK_FUNC) {
			if (f->type == FAWK_NIL)
				goto setup; /* forward call - sym is in the symtab as NIL */
			else
				FAWK_ERROR(ctx, 64+strlen(name), (FAWK_ERR, "funcreg: '%s' collides with another global symbol\n", name));
		}
		if ((f->data.func.ip == FAWK_CODE_INVALID) && (addr != FAWK_CODE_INVALID)) {
			f->data.func.ip = addr;
			f->data.func.numargs = numargs;
		}
	}
	return 0;
}

FAWK_API fawk_cell_t *fawk_symtab_regvar(fawk_ctx_t *ctx, const char *name, fawk_celltype_t tclass)
{
	fawk_cell_t *f = htsp_get(&ctx->symtab, name);

	fawk_assert((tclass == FAWK_SCALAR) || (tclass == FAWK_ARRAY)); /* functions are registered by fawk_symtab_regfunc() */

	if (f == NULL) {
		FAWK_SYMTAB_REG(tclass, name, NULL, NULL);
		if (tclass == FAWK_ARRAY)
			fawk_array_init(ctx, f);
	}
	/* else :  accept existing entry if it was a forward declaration */
	return f;
}
